#!/usr/bin/env python3

SUPPORTED_LANGUAGES = ["ar-EG", "ca-ES", "cs-CZ", "da-DK", "de-DE", "en-GB", "en-US", "eo-ZZ",\
                       "es-ES", "fi-FI", "fr-CA", "fr-FR", "gl-ES", "hu-HU", "it-IT", "ja-JP",\
                       "ko-KR", "nb-NO", "nl-NL", "pl-PL", "pt-BR", "ru-RU", "sv-SE", "tr-TR",\
                       "uk-UA", "vi-VN", "zh-CN", "zh-TW"]



#!/usr/bin/env python3
"""
Dump translations for OpenRCT2's JSON objects given a language and a file
"""

import argparse
import glob
import json
import os
import sys
# from pprint import pprint

from languages import SUPPORTED_LANGUAGES

def dir_path(string):
    """ Checks for a valid dir_path """
    if os.path.isdir(string):
        return string
    raise NotADirectoryError(string)

def get_arg_parser():
    """ Command line arguments """
    parser = argparse.ArgumentParser(description='Dump translations for OpenRCT2\'s JSON objects.',\
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-o', '--objects', default="objects", help='JSON objects directory')
    parser.add_argument('-f', '--fallback', default="en-GB",\
                        help='Fallback language to check against', choices=SUPPORTED_LANGUAGES)
    target_group = parser.add_mutually_exclusive_group(required=True)
    target_group.add_argument('-t', '--target_dir', help='Directory to dump translation files',\
                              type=dir_path)
    target_group.add_argument('-d', '--dumpfile', help='Translation file to export to')
    language_group = parser.add_mutually_exclusive_group(required=True)
    language_group.add_argument('-l', '--language', choices=SUPPORTED_LANGUAGES,\
                                help='Language that should be extracted, e.g. ja-JP')
    language_group.add_argument('-a', '--all-languages', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,\
                        help='Maximize information printed on screen')
    parser.add_argument('-m', '--missing', action='store_true', default=False,\
                        help='Only dump keys with missing translation')
    return parser

def parse_required_switch_pairs(args):
    """ Make sure only valid switch pairs are used """
    single_language = args.language and args.dumpfile
    all_languages = args.all_languages and args.target_dir
    if not single_language and not all_languages:
        print(f"Invalid switch pair. Use '-l <lang> -d <file>' or '-a -t <target_dir>")
        sys.exit()

def add_key_value(strings_by_object, obj_id, key, value):
    """ Add key-value pair to id entry in dict, initializing id if not valid """
    if not obj_id in strings_by_object:
        strings_by_object[obj_id] = {}
    strings_by_object[obj_id][key] = value

def extract_translations(language_to_extract, args, filename, strings_by_object):
    """ Read JSON file and extracts translations for a given language """

    fallback_language = args.fallback
    verbose = args.verbose
    reference_str_count = 0
    translated_str_count = 0

    with open(filename, encoding="utf8") as file:
        data = json.load(file)

        if not 'strings' in data:
            print(f"No strings in {data['id']}, skipping")
            return (translated_str_count, reference_str_count)

        for string_key in data['strings']:
            if fallback_language in data['strings'][string_key]:
                fallback_translation = data['strings'][string_key][fallback_language]
            else:
                fallback_translation = ""

            if language_to_extract in data['strings'][string_key]:
                if verbose:
                    print(f"Found existing translation for {data['id']}")
                current_translation = data['strings'][string_key][language_to_extract]
                ref_key = f'reference-{string_key}'
                if not args.missing:
                    add_key_value(strings_by_object, data['id'], ref_key, fallback_translation)
                    add_key_value(strings_by_object, data['id'], string_key, current_translation)
                reference_str_count += 1
                translated_str_count += 1
            elif fallback_language in data['strings'][string_key]:
                if verbose:
                    print(f"No existing translation for {data['id']} yet,"
                          f" using {fallback_language}")
                ref_key = f'reference-{string_key}'
                add_key_value(strings_by_object, data['id'], ref_key, fallback_translation)
                add_key_value(strings_by_object, data['id'], string_key, fallback_translation)
                reference_str_count += 1
            else:
                if verbose:
                    print(f"No existing translation for {data['id']} yet,"
                          f" but no {fallback_language} string either -- skipping")
    return (translated_str_count, reference_str_count)

def dump_translation(language_to_extract, args, dump_file_name):
    """ Dump a language translation for OpenRCT2's JSON objects """
    strings_by_object = {}
    reference_str_count = 0
    translated_str_count = 0

    for filename in glob.iglob(args.objects + '/**/*.json', recursive=True):
        obj_translations, ref_obj_translation =\
            extract_translations(language_to_extract, args, filename, strings_by_object)
        reference_str_count += ref_obj_translation
        translated_str_count += obj_translations

    translation_progress = round(100 * translated_str_count / reference_str_count, 2)
    print(f'{language_to_extract}: {translation_progress}% completeness')

    # Sort output by object ids (dict keys)
    strings_by_object = dict(sorted(strings_by_object.items()))

    out = open(dump_file_name, "w", encoding="utf8")
    json.dump(strings_by_object, out, indent=4, ensure_ascii=False, separators=(',', ': '))
    out.write("\n")
    out.close()

def dump_translations():
    """ Dump translations for OpenRCT2's JSON objects """
    parser = get_arg_parser()
    args = parser.parse_args()
    parse_required_switch_pairs(args)
    languages_to_extract = SUPPORTED_LANGUAGES if args.all_languages else [args.language]
    for lang in languages_to_extract:
        dump_file_name = f'{args.target_dir}/{lang}.json' if args.all_languages else args.dumpfile
        dump_translation(lang, args, dump_file_name)

if __name__ == "__main__":
    dump_translations()



#!/usr/bin/env python3
"""
Imports translations into OpenRCT2's JSON objects given a language and an input file
"""

import argparse
import glob
import json
import os
import re
import sys

from languages import SUPPORTED_LANGUAGES

def dir_path(string):
    """ Checks for a valid dir_path """
    if os.path.isdir(string):
        return string
    raise NotADirectoryError(string)

def get_arg_parser():
    """ Command line arguments """
    parser = argparse.ArgumentParser(description=\
                                     'Imports translations into OpenRCT2\'s JSON objects.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-o', '--objects', default="objects", help='JSON objects directory')
    parser.add_argument('-f', '--fallback', default="en-GB",\
                        help='Fallback language to check against', choices=SUPPORTED_LANGUAGES)
    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument('-i', '--input', help='Translation dump file to import from')
    input_group.add_argument('-d', '--dir', type=dir_path,\
                             help='Directory with translation dump files to import from')
    language_group = parser.add_mutually_exclusive_group(required=True)
    language_group.add_argument('-l', '--language', choices=SUPPORTED_LANGUAGES,\
                                help='Language that is being translated, e.g. ja-JP')
    language_group.add_argument('-a', '--all-languages', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true', default=False,\
                        help='Maximize information printed on screen')
    return parser

def parse_required_switch_pairs(args):
    """ Make sure only valid switch pairs are used """
    single_language = args.language and args.input
    all_languages = args.all_languages and args.dir
    if not single_language and not all_languages:
        print(f"Invalid switch pair. Use '-l <lang> -i <file>' or '-a -d <target_dir>")
        sys.exit()

class LessVerboseJSONEncoder(json.JSONEncoder):
    """ Custom JSON Encoder that reduces output verbosity """
    def iterencode(self, o, _one_shot=False):
        list_lvl = 0
        for string in super(LessVerboseJSONEncoder, self).iterencode(o, _one_shot=_one_shot):
            if string.startswith('['):
                list_lvl += 1
                string = re.sub(r'\n\s*', '', string).strip()
            elif list_lvl > 0:
                string = re.sub(r'\n\s*', ' ', string).strip()
                if string and string[-1] == ',':
                    string = string[:-1] + self.item_separator
                elif string and string[-1] == ':':
                    string = string[:-1] + self.key_separator
            if string.endswith(']'):
                list_lvl -= 1
            yield string

def update_object_translation(filename, data, file):
    """ Update target object translation """
    file = open(filename, "w", encoding="utf8")
    json.dump(data, file, indent=4, separators=(',', ': '),
              ensure_ascii=False, cls=LessVerboseJSONEncoder)
    file.write("\n")
    file.close()

def translatable(verbose, data):
    """ Checks if the file has string content """
    if 'strings' in data:
        return True
    if verbose:
        print(f"No strings in {data['id']} -- skipping")
    return False

def is_object_translated(verbose, object_id, strings_by_object):
    """ Checks if there are any translations for the given file id """
    if object_id in strings_by_object:
        return True
    if verbose:
        print(f"No translations for {object_id} in dump file -- skipping")
    return False

def is_key_translated(verbose, obj_id, key, object_json):
    """ Checks if the given key of an object is translated in input JSON """
    if key in object_json:
        return True
    if verbose:
        print(f"No translation for {obj_id} string '{key}' in dump file -- skipping")
    return False

def fallback_key_exists(verbose, fallback_language, obj_id, key, object_json):
    """ Checks if there was source language to be translated from """
    if fallback_language in object_json:
        return True
    if verbose:
        print(f"No {fallback_language} reference for {obj_id} string '{key}'"
              f" in dump file -- probably shouldn't exist; skipping")
    return False

def translation_existed(target_lang, translations):
    """ Checks if the translation for a given key existed """
    if target_lang in translations:
        return True
    return False

def translation_changed(verbose, obj_id, key, translation, target_string):
    """ Checks if the translation for a given key changed """
    if target_string == translation:
        if verbose:
            print(f"Translation for {obj_id} string '{key}' has not changed -- skipping")
        return False
    return True

def update_translation(target_lang, ref_lang, verbose, filename, strings_by_object):
    """ Updates an object translation, if applicable """
    file = open(filename, encoding="utf8")
    data = json.load(file)
    file.close()
    obj_id = data['id']

    if not translatable(verbose, data):
        return
    if not is_object_translated(verbose, obj_id, strings_by_object):
        return

    updated = False
    for string_key in data['strings']:

        if not is_key_translated(verbose, obj_id, string_key, strings_by_object[obj_id]):
            continue

        if not fallback_key_exists(verbose, ref_lang, string_key, obj_id,
                                   data['strings'][string_key]):
            continue

        previously_translated = translation_existed(target_lang, data['strings'][string_key])
        if previously_translated:
            translation = data['strings'][string_key][target_lang]
        else:
            translation = data['strings'][string_key][ref_lang]
        target_string = strings_by_object[obj_id][string_key]
        if translation_changed(verbose, obj_id, string_key, translation, target_string):
            print(f"{target_lang}: Updating {obj_id} string '{string_key}'")
            data['strings'][string_key][target_lang] = strings_by_object[obj_id][string_key]
            updated = True

    if updated:
        update_object_translation(filename, data, file)

def load_translation(language, fallback_language, verbose, input_filename, objects):
    """ Load translation from the given file into each object JSON """
    in_file = open(input_filename, encoding="utf8")
    strings_by_object = json.load(in_file)
    in_file.close()

    for filename in glob.iglob(objects + '/**/*.json', recursive=True):
        update_translation(language, fallback_language, verbose, filename, strings_by_object)

def load_translations():
    """ Load translations from the given files into each object JSON """
    parser = get_arg_parser()
    args = parser.parse_args()
    parse_required_switch_pairs(args)
    languages_to_extract = SUPPORTED_LANGUAGES if args.all_languages else [args.language]
    for lang in languages_to_extract:
        read_file_name = f'{args.dir}/{lang}.json' if args.all_languages else args.input
        load_translation(lang, args.fallback, args.verbose, read_file_name, args.objects)

if __name__ == "__main__":
    load_translations()



#!/usr/bin/env python3

from languages import SUPPORTED_LANGUAGES
from unidiff import PatchSet
from unidiff.constants import LINE_TYPE_EMPTY

import argparse
import os

def dir_path(string):
    """ Checks for a valid dir_path """
    if os.path.isdir(string):
        return string
    raise NotADirectoryError(string)

def get_arg_parser():
    """ Command line arguments """
    parser = argparse.ArgumentParser(description=\
                                     'Cleans up a patch file to apply to the objects repository.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-p', '--patch', required=True, help='Path to the patch file to clean up.')
    parser.add_argument('-l', '--language', required=True, choices=SUPPORTED_LANGUAGES,\
                                help='Language that is being translated, e.g. ja-JP')
    return parser

class PatchCleaner:
    """
    Cleans a given unified diff such that only lines matching the language parameter are included,
    as well as any preceding lines that may be giving way to them (e.g. by adding a trailing comma).
    """

    def __init__(self, filename, language):
        self.patch = PatchSet.from_filename(filename)
        self.language = language
        self.clean_patch()

    def __str__(self):
        return str(self.patch)

    def clean_patch(self):
        """ Cleans the patch, removing irrelevant files. """

        files_to_be_removed = []
        for i, file in enumerate(self.patch):
            self.clean_file(file)

            if not file.added and not file.removed:
                files_to_be_removed.append(i)

        # Any files to be removed?
        # (In reverse order, as we'll be reindexing.)
        if len(files_to_be_removed):
            for i in reversed(files_to_be_removed):
                del self.patch[i]

    def clean_file(self, file):
        """ Cleans one particular file in the patch set, removing empty hunks. """

        hunks_to_be_removed = []
        for j, hunk in enumerate(file):
            self.clean_hunk(hunk)

            # Is this hunk still modifying anything?
            # If not, we'll drop it after iterating everything.
            if not hunk.added and not hunk.removed:
                hunks_to_be_removed.append(j)

        # Any hunks to be removed?
        # (In reverse order, as we'll be reindexing.)
        if len(hunks_to_be_removed):
            for j in reversed(hunks_to_be_removed):
                del file[j]

    def clean_hunk(self, hunk):
        """ Cleans one particular hunk in the patch set, removing irrelevant lines. """

        for k, line in enumerate(hunk):
            # Is this line modifying anything?
            if line.is_added or line.is_removed:
                # We're definitely keeping lines that have something to do
                # with the target language.
                if self.language in line.value:
                    continue

                # Is this line a removal followed by an addition of the same language,
                # in turn followed by the target language?
                if self.is_accommodating_change(hunk, k):
                    # TODO: Check languages match -- good enough a heuristic for now.
                    continue

                # Otherwise, exclude this change from the patch.
                else:
                    line.line_type = LINE_TYPE_EMPTY

    def is_accommodating_change(self, hunk, k):
        """
        Determines whether the line current line 'k' is part of a change that's
        accommodating the target language in some way. For example, the addition
        of a trailing comma in a seemingly unrelated line:

        -"en-GB": "Penguin Trains"
        +"en-GB": "Penguin Trains",
        +"ja-JP": "ペンギンボブスレー"
        """

        # Is line k a removal followed by an addition of the same language,
        # in turn followed by the target language?
        if hunk[k].is_removed and hunk[k + 1].is_added and \
            self.language in hunk[k + 2].value:
            return True

        # Indeed, is the previous line a removal and the next line an addition
        # of the target language?
        elif hunk[k - 1].is_removed and hunk[k].is_added and \
            self.language in hunk[k + 1].value:
            return True

        return False


if __name__ == "__main__":
    parser = get_arg_parser()
    args = parser.parse_args()

    patch = PatchCleaner(args.patch, args.language)
    print(patch)
